import re
import grpc
import json
import collections
from git import Repo
from functools import lru_cache
from typing import Union, Dict, Any

from nbox.auth import inside_pod, secret
from nbox.init import MetadataInjectInterceptor, WorkspaceIdInjectInterceptor
from nbox.utils import logger, lo
from nbox import messages as mpb
from nbox.hyperloop.common.common_pb2 import Resource

from nbox.lmao_v4.proto.lmao_service_pb2_grpc import LMAOStub
from nbox.lmao_v4.proto.logs_pb2 import TrackerLog
from nbox.lmao_v4.proto.project_pb2 import Project as ProjectProto

@lru_cache(1)
def get_lmao_stub() -> LMAOStub:
  token_cred = grpc.access_token_call_credentials(secret.access_token)
  ssl_creds = grpc.ssl_channel_credentials()
  creds = grpc.composite_channel_credentials(ssl_creds, token_cred)
  channel = grpc.secure_channel(secret.nbx_url.replace("https://", "dns:/") + ":443", creds)

  # channel = grpc.insecure_channel("dns:/0.0.0.0:50051") # debug mode

  future = grpc.channel_ready_future(channel)
  future.add_done_callback(lambda _: logger.debug(f"NBX 'LMAO' gRPC stub is ready!"))
  interceptors = [
    MetadataInjectInterceptor(),
    WorkspaceIdInjectInterceptor(),
  ]
  channel = grpc.intercept_channel(channel, *interceptors)
  stub = LMAOStub(channel)
  return stub

valid_key_regex = re.compile(r'[a-zA-Z0-9_\-\.]+$') 

def flatten(d, parent_key='', sep='.'):
  """flatten a dictionary with prefix and seperator, if """
  items = []
  for k, v in d.items():
    if sep in k:
      raise ValueError(f"Key '{k}' contains the seperator '{sep}', this can cause issue during rebuilding")
    new_key = parent_key + sep + k if parent_key else k
    if isinstance(v, collections.abc.MutableMapping):
      items.extend(flatten(v, new_key, sep=sep).items())
    else:
      items.append((new_key, v))
  return dict(items)


def get_tracker_log(log: dict) -> TrackerLog:
  items = flatten(log).items()
  out = TrackerLog()
  for k,v in items:
    if valid_key_regex.match(k) is None:
      raise ValueError(f"Invalid key '{k}', allowed letters digits and '_-'")
    if type(v) in (float, int):
      out.number_keys.append(k)
      out.number_values.append(v)
    elif type(v) == str:
      out.text_keys.append(k)
      out.text_values.append(v)
    else:
      raise ValueError(f"Invalid type '{type(v)}' for key '{k}'")
  return out


# there are some legacy functions that were built for lmao_v2, in some cases retrofitted for v4

def get_git_details(folder):
  """If there is a `.git` folder in the folder, return some details for that."""
  repo = Repo(folder)

  # check for any unstaged files
  uncommited_files = {}
  diff = repo.index.diff(None)
  for f in diff:
    path = f.a_path or f.b_path # when new file is added, a_path is None
    uncommited_files[path] = f.change_type
  if uncommited_files:
    logger.warning(f"Uncommited files: {uncommited_files}")

  # get the remote url
  try:
    remote_url = repo.remote().url

    # clean the remote_url because it can sometimes contain the repo token as well.
    # this can become security hazard. so if you have an example, that is not suppported:
    #   go ahead, make a PR!
    if "github.com" in remote_url:
      remote_url = re.sub(r"ghp_\w+@", "", remote_url)
  except ValueError:
    remote_url = None

  # get the size of the repository
  size = None
  for line in repo.git.count_objects("-v").splitlines():
    if line.startswith("size:"):
      size = int(line[len("size:") :].strip())
  if size > (1 << 30):
    logger.warning(f"Repository size over 1GB, you might want to work on it")

  return {
    "remote_url": remote_url,
    "branch": repo.active_branch.name,
    "commit": repo.head.commit.hexsha,
    "uncommited_files": uncommited_files,
    "untracked_files": repo.untracked_files,
    "size": size,
  }

def get_project(project_id: str) -> ProjectProto:
  lmao_stub = get_lmao_stub()
  project: ProjectProto = lmao_stub.GetProject(ProjectProto(id=project_id))
  return project


def resource_from_dict(d: Dict[str, Any]):
  return Resource(
    cpu = str(d.get("cpu", "")),
    memory = str(d.get("memory", "")),
    disk_size = str(d.get("disk_size", "")),
    gpu = str(d.get("gpu", "")),
    gpu_count = str(d.get("gpu_count", "")),
    timeout = int(d.get("timeout", 0)),
    max_retries = int(d.get("max_retries", 0)),
  )



"""
Common structs
"""


class ExperimentConfig:
  def __init__(
    self,
    run_kwargs: Dict[str, Any],
    git: Dict[str, Any],
    resource: Resource,
    cli_comm: str,
    save_to_relic: bool = True,
    enable_system_monitoring: bool = False,
  ):
    """In an ideal world this would be a protobuf message, but we are not there yet. This contains all the
    things that are stored in the DB and FE can use this to render elements in the UI.

    Args:
      run_kwargs (Dict[str, Any]): All the arguments that the user passed to the `nbx lmao run ...` CLI
      git (Dict[str, Any]): details for the git repo
      resource (Resource): Resource pb object that contains the details of the resource
      cli_comm (str): The CLI command that was used to run the experiment, user can use this to reproduce the experiment
      save_to_relic (bool): If the user wants to save the experiment to the relic by default
      enable_system_monitoring (bool): If the user wants to enable system monitoring
    """
    self.run_kwargs = run_kwargs
    self.git = git
    self.resource = resource
    self.cli_comm = cli_comm
    self.save_to_relic = save_to_relic
    self.enable_system_monitoring = enable_system_monitoring

  def to_dict(self):
    return {
      "run_kwargs": self.run_kwargs,
      "git": self.git,
      "resource": mpb.message_to_dict(self.resource),
      "cli_comm": self.cli_comm,
      "save_to_relic": self.save_to_relic,
      "enable_system_monitoring": self.enable_system_monitoring,
    }
  
  @classmethod
  def from_dict(cls, data):
    if not isinstance(data, Resource):
      data["resource"] = resource_from_dict(data["resource"])
    return cls(**data)

  def to_json(self):
    return json.dumps(self.to_dict())

  @classmethod
  def from_json(cls, json_str):
    d = json.loads(json_str)
    d["resource"] = resource_from_dict(d["resource"])
    return cls(**d)



class LiveConfig:
  def __init__(
    self,
    resource: Resource,
    cli_comm: str,
    enable_system_monitoring: bool = False,
    extra_kwargs: Dict[str, Any] = {},
  ):
    self.resource = resource
    self.cli_comm = cli_comm
    self.enable_system_monitoring = enable_system_monitoring
    self.extra_kwargs = extra_kwargs

  def to_dict(self):
    return {
      "resource": mpb.message_to_dict(self.resource),
      "cli_comm": self.cli_comm,
      "enable_system_monitoring": self.enable_system_monitoring,
      "extra_kwargs": self.extra_kwargs,
    }
  
  def to_json(self):
    return json.dumps(self.to_dict())
  
  @classmethod
  def from_json(cls, json_str) -> 'LiveConfig':
    d = json.loads(json_str)
    d["resource"] = resource_from_dict(d["resource"])
    _cls = cls(**d)
    return _cls


"""
Constants
"""

# do not change these it can become a huge pain later on
LMAO_RELIC_NAME = "experiments"
LMAO_RM_PREFIX = "NBXLmao-"
LMAO_SERVING_FILE = "NBXLmaoServingCfg.json"
LMAO_ENV_VAR_PREFIX = "NBX_LMAO_"



from nbox.lmao_v4.common import get_lmao_stub, get_git_details, get_project, get_tracker_log, LMAO_RM_PREFIX, ExperimentConfig
from nbox.lmao_v4.tracker import Tracker
from nbox.lmao_v4.cli import LmaoCLI


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lmao_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
try:
  from gen.proto import logs_pb2 as logs__pb2
except ImportError:
  from nbox.lmao_v4.proto import logs_pb2 as logs__pb2
try:
  from gen.proto import project_pb2 as project__pb2
except ImportError:
  from nbox.lmao_v4.proto import project_pb2 as project__pb2
try:
  from gen.proto import rules_pb2 as rules__pb2
except ImportError:
  from nbox.lmao_v4.proto import rules_pb2 as rules__pb2
try:
  from gen.proto import tracker_pb2 as tracker__pb2
except ImportError:
  from nbox.lmao_v4.proto import tracker_pb2 as tracker__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12lmao_service.proto\x12\x07lmao_pb\x1a\x1cgoogle/api/annotations.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\nlogs.proto\x1a\rproject.proto\x1a\x0brules.proto\x1a\rtracker.proto2\xa9\x10\n\x04LMAO\x12m\n\x0bInitProject\x12\x1b.lmao_pb.InitProjectRequest\x1a\x1c.lmao_pb.InitProjectResponse\"#\x82\xd3\xe4\x93\x02\x1d\"\x18/monitoring/init_project:\x01*\x12T\n\nGetProject\x12\x10.lmao_pb.Project\x1a\x10.lmao_pb.Project\"\"\x82\xd3\xe4\x93\x02\x1c\"\x17/monitoring/get_project:\x01*\x12Z\n\rUpdateProject\x12\x10.lmao_pb.Project\x1a\x10.lmao_pb.Project\"%\x82\xd3\xe4\x93\x02\x1f\"\x1a/monitoring/update_project:\x01*\x12`\n\rDeleteProject\x12\x10.lmao_pb.Project\x1a\x16.google.protobuf.Empty\"%\x82\xd3\xe4\x93\x02\x1f\"\x1a/monitoring/delete_project:\x01*\x12q\n\x0cListProjects\x12\x1c.lmao_pb.ListProjectsRequest\x1a\x1d.lmao_pb.ListProjectsResponse\"$\x82\xd3\xe4\x93\x02\x1e\"\x19/monitoring/list_projects:\x01*\x12\x65\n\x0eGetRuleBuilder\x12\x14.lmao_pb.RuleBuilder\x1a\x14.lmao_pb.RuleBuilder\"\'\x82\xd3\xe4\x93\x02!\"\x1c/monitoring/get_rule_builder:\x01*\x12Y\n\nCreateRule\x12\x18.lmao_pb.InitRuleRequest\x1a\r.lmao_pb.Rule\"\"\x82\xd3\xe4\x93\x02\x1c\"\x17/monitoring/create_rule:\x01*\x12H\n\x07GetRule\x12\r.lmao_pb.Rule\x1a\r.lmao_pb.Rule\"\x1f\x82\xd3\xe4\x93\x02\x19\"\x14/monitoring/get_rule:\x01*\x12N\n\nUpdateRule\x12\r.lmao_pb.Rule\x1a\r.lmao_pb.Rule\"\"\x82\xd3\xe4\x93\x02\x1c\"\x17/monitoring/update_rule:\x01*\x12W\n\nDeleteRule\x12\r.lmao_pb.Rule\x1a\x16.google.protobuf.Empty\"\"\x82\xd3\xe4\x93\x02\x1c\"\x17/monitoring/delete_rule:\x01*\x12V\n\tListRules\x12\x12.lmao_pb.RulesList\x1a\x12.lmao_pb.RulesList\"!\x82\xd3\xe4\x93\x02\x1b\"\x16/monitoring/list_rules:\x01*\x12\x61\n\x0bInitTracker\x12\x1b.lmao_pb.InitTrackerRequest\x1a\x10.lmao_pb.Tracker\"#\x82\xd3\xe4\x93\x02\x1d\"\x18/monitoring/init_tracker:\x01*\x12T\n\nGetTracker\x12\x10.lmao_pb.Tracker\x1a\x10.lmao_pb.Tracker\"\"\x82\xd3\xe4\x93\x02\x1c\"\x17/monitoring/get_tracker:\x01*\x12Z\n\rUpdateTracker\x12\x10.lmao_pb.Tracker\x1a\x10.lmao_pb.Tracker\"%\x82\xd3\xe4\x93\x02\x1f\"\x1a/monitoring/update_tracker:\x01*\x12`\n\rDeleteTracker\x12\x10.lmao_pb.Tracker\x1a\x16.google.protobuf.Empty\"%\x82\xd3\xe4\x93\x02\x1f\"\x1a/monitoring/delete_tracker:\x01*\x12q\n\x0cListTrackers\x12\x1c.lmao_pb.ListTrackersRequest\x1a\x1d.lmao_pb.ListTrackersResponse\"$\x82\xd3\xe4\x93\x02\x1e\"\x19/monitoring/list_trackers:\x01*\x12s\n\x0fGetTrackerTable\x12\x1f.lmao_pb.GetTrackerTableRequest\x1a\x15.lmao_pb.TrackerTable\"(\x82\xd3\xe4\x93\x02\"\"\x1d/monitoring/get_tracker_table:\x01*\x12\x63\n\rPutTrackerLog\x12\x13.lmao_pb.TrackerLog\x1a\x15.lmao_pb.TrackerLogId\"&\x82\xd3\xe4\x93\x02 \"\x1b/monitoring/put_tracker_log:\x01*\x12r\n\x0eGetTrackerLogs\x12\x1a.lmao_pb.TrackerLogRequest\x1a\x1b.lmao_pb.TrackerLogResponse\"\'\x82\xd3\xe4\x93\x02!\"\x1c/monitoring/get_tracker_logs:\x01*\x12o\n\rCreateDataset\x12\x1e.lmao_pb.TrackerDatasetRequest\x1a\x17.lmao_pb.TrackerDataset\"%\x82\xd3\xe4\x93\x02\x1f\"\x1a/monitoring/create_dataset:\x01*\x12u\n\x13\x43reateDatasetStatus\x12\x17.lmao_pb.TrackerDataset\x1a\x17.lmao_pb.TrackerDataset\",\x82\xd3\xe4\x93\x02&\"!/monitoring/create_dataset_status:\x01*B/Z-github.com/NimbleBoxAI/nimblebox-lmao/lmao_pbb\x06proto3')



_LMAO = DESCRIPTOR.services_by_name['LMAO']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z-github.com/NimbleBoxAI/nimblebox-lmao/lmao_pb'
  _LMAO.methods_by_name['InitProject']._options = None
  _LMAO.methods_by_name['InitProject']._serialized_options = b'\202\323\344\223\002\035\"\030/monitoring/init_project:\001*'
  _LMAO.methods_by_name['GetProject']._options = None
  _LMAO.methods_by_name['GetProject']._serialized_options = b'\202\323\344\223\002\034\"\027/monitoring/get_project:\001*'
  _LMAO.methods_by_name['UpdateProject']._options = None
  _LMAO.methods_by_name['UpdateProject']._serialized_options = b'\202\323\344\223\002\037\"\032/monitoring/update_project:\001*'
  _LMAO.methods_by_name['DeleteProject']._options = None
  _LMAO.methods_by_name['DeleteProject']._serialized_options = b'\202\323\344\223\002\037\"\032/monitoring/delete_project:\001*'
  _LMAO.methods_by_name['ListProjects']._options = None
  _LMAO.methods_by_name['ListProjects']._serialized_options = b'\202\323\344\223\002\036\"\031/monitoring/list_projects:\001*'
  _LMAO.methods_by_name['GetRuleBuilder']._options = None
  _LMAO.methods_by_name['GetRuleBuilder']._serialized_options = b'\202\323\344\223\002!\"\034/monitoring/get_rule_builder:\001*'
  _LMAO.methods_by_name['CreateRule']._options = None
  _LMAO.methods_by_name['CreateRule']._serialized_options = b'\202\323\344\223\002\034\"\027/monitoring/create_rule:\001*'
  _LMAO.methods_by_name['GetRule']._options = None
  _LMAO.methods_by_name['GetRule']._serialized_options = b'\202\323\344\223\002\031\"\024/monitoring/get_rule:\001*'
  _LMAO.methods_by_name['UpdateRule']._options = None
  _LMAO.methods_by_name['UpdateRule']._serialized_options = b'\202\323\344\223\002\034\"\027/monitoring/update_rule:\001*'
  _LMAO.methods_by_name['DeleteRule']._options = None
  _LMAO.methods_by_name['DeleteRule']._serialized_options = b'\202\323\344\223\002\034\"\027/monitoring/delete_rule:\001*'
  _LMAO.methods_by_name['ListRules']._options = None
  _LMAO.methods_by_name['ListRules']._serialized_options = b'\202\323\344\223\002\033\"\026/monitoring/list_rules:\001*'
  _LMAO.methods_by_name['InitTracker']._options = None
  _LMAO.methods_by_name['InitTracker']._serialized_options = b'\202\323\344\223\002\035\"\030/monitoring/init_tracker:\001*'
  _LMAO.methods_by_name['GetTracker']._options = None
  _LMAO.methods_by_name['GetTracker']._serialized_options = b'\202\323\344\223\002\034\"\027/monitoring/get_tracker:\001*'
  _LMAO.methods_by_name['UpdateTracker']._options = None
  _LMAO.methods_by_name['UpdateTracker']._serialized_options = b'\202\323\344\223\002\037\"\032/monitoring/update_tracker:\001*'
  _LMAO.methods_by_name['DeleteTracker']._options = None
  _LMAO.methods_by_name['DeleteTracker']._serialized_options = b'\202\323\344\223\002\037\"\032/monitoring/delete_tracker:\001*'
  _LMAO.methods_by_name['ListTrackers']._options = None
  _LMAO.methods_by_name['ListTrackers']._serialized_options = b'\202\323\344\223\002\036\"\031/monitoring/list_trackers:\001*'
  _LMAO.methods_by_name['GetTrackerTable']._options = None
  _LMAO.methods_by_name['GetTrackerTable']._serialized_options = b'\202\323\344\223\002\"\"\035/monitoring/get_tracker_table:\001*'
  _LMAO.methods_by_name['PutTrackerLog']._options = None
  _LMAO.methods_by_name['PutTrackerLog']._serialized_options = b'\202\323\344\223\002 \"\033/monitoring/put_tracker_log:\001*'
  _LMAO.methods_by_name['GetTrackerLogs']._options = None
  _LMAO.methods_by_name['GetTrackerLogs']._serialized_options = b'\202\323\344\223\002!\"\034/monitoring/get_tracker_logs:\001*'
  _LMAO.methods_by_name['CreateDataset']._options = None
  _LMAO.methods_by_name['CreateDataset']._serialized_options = b'\202\323\344\223\002\037\"\032/monitoring/create_dataset:\001*'
  _LMAO.methods_by_name['CreateDatasetStatus']._options = None
  _LMAO.methods_by_name['CreateDatasetStatus']._serialized_options = b'\202\323\344\223\002&\"!/monitoring/create_dataset_status:\001*'
  _LMAO._serialized_start=146
  _LMAO._serialized_end=2235
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: logs.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
try:
  from gen.proto import tracker_pb2 as tracker__pb2
except ImportError:
  from nbox.lmao_v4.proto import tracker_pb2 as tracker__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\nlogs.proto\x12\x07lmao_pb\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/wrappers.proto\x1a\rtracker.proto\"\xdd\x01\n\nTrackerLog\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x12\n\ntracker_id\x18\x03 \x01(\t\x12\x13\n\x0bnumber_keys\x18\x06 \x03(\t\x12\x15\n\rnumber_values\x18\x07 \x03(\x02\x12\x11\n\ttext_keys\x18\x08 \x03(\t\x12\x13\n\x0btext_values\x18\t \x03(\t\x12-\n\ttimestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06log_id\x18\x05 \x01(\t\"\x1e\n\x0cTrackerLogId\x12\x0e\n\x06log_id\x18\x01 \x01(\t\"\xa4\x02\n\x0cRecordColumn\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\nvalue_type\x18\x02 \x01(\x0e\x32\x1e.lmao_pb.RecordColumn.DataType\x12.\n\x04rows\x18\x03 \x03(\x0b\x32 .lmao_pb.RecordColumn.RecordItem\x1at\n\nRecordItem\x12\t\n\x01x\x18\x01 \x01(\x03\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06number\x18\x03 \x01(\x02\x12\x0c\n\x04text\x18\x04 \x01(\t\x12\x0e\n\x06log_id\x18\x05 \x01(\t\"-\n\x08\x44\x61taType\x12\t\n\x05UNSET\x10\x00\x12\n\n\x06NUMBER\x10\x01\x12\n\n\x06STRING\x10\x02\"G\n\tRecordRow\x12\x12\n\ntracker_id\x18\x01 \x01(\t\x12&\n\x07\x63olumns\x18\x02 \x03(\x0b\x32\x15.lmao_pb.RecordColumn\"\xd6\x01\n\x11TrackerLogRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x12\n\ntracker_id\x18\x03 \x01(\t\x12\x0c\n\x04keys\x18\x04 \x03(\t\x12)\n\x05\x61\x66ter\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12*\n\x06\x62\x65\x66ore\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\r\n\x05limit\x18\x07 \x01(\x05\x12\x0f\n\x07log_ids\x18\x08 \x03(\t\"9\n\x12TrackerLogResponse\x12#\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32\x15.lmao_pb.RecordColumn\"\x8d\x01\n\x16GetTrackerTableRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12*\n\x0ctracker_type\x18\x03 \x01(\x0e\x32\x14.lmao_pb.TrackerType\x12\x0e\n\x06offset\x18\x04 \x01(\x05\x12\r\n\x05limit\x18\x05 \x01(\x05\"7\n\x0cTrackerTable\x12\'\n\x0brecord_rows\x18\x01 \x03(\x0b\x32\x12.lmao_pb.RecordRow\"\x95\x03\n\x15TrackerDatasetRequest\x12\x11\n\trelics_id\x18\x01 \x01(\t\x12\x13\n\x0brelics_path\x18\x02 \x01(\t\x12\x42\n\rexport_format\x18\x03 \x01(\x0e\x32+.lmao_pb.TrackerDatasetRequest.ExportFormat\x12\x31\n\x0blog_request\x18\n \x01(\x0b\x32\x1a.lmao_pb.TrackerLogRequestH\x00\x12(\n\x05query\x18\x0b \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x12\x38\n\x07log_ids\x18\x0c \x01(\x0b\x32%.lmao_pb.TrackerDatasetRequest.LogIdsH\x00\x12&\n\x05\x65xtra\x18\x14 \x01(\x0b\x32\x17.google.protobuf.Struct\x1a\x19\n\x06LogIds\x12\x0f\n\x07log_ids\x18\x01 \x03(\t\"\'\n\x0c\x45xportFormat\x12\x0c\n\x08UNSET_EF\x10\x00\x12\t\n\x05JSONL\x10\x01\x42\r\n\x0bquery_style\"S\n\x0eTrackerDataset\x12\x12\n\ndataset_id\x18\x01 \x01(\t\x12-\n\tcompleted\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.BoolValueB/Z-github.com/NimbleBoxAI/nimblebox-lmao/lmao_pbb\x06proto3')



_TRACKERLOG = DESCRIPTOR.message_types_by_name['TrackerLog']
_TRACKERLOGID = DESCRIPTOR.message_types_by_name['TrackerLogId']
_RECORDCOLUMN = DESCRIPTOR.message_types_by_name['RecordColumn']
_RECORDCOLUMN_RECORDITEM = _RECORDCOLUMN.nested_types_by_name['RecordItem']
_RECORDROW = DESCRIPTOR.message_types_by_name['RecordRow']
_TRACKERLOGREQUEST = DESCRIPTOR.message_types_by_name['TrackerLogRequest']
_TRACKERLOGRESPONSE = DESCRIPTOR.message_types_by_name['TrackerLogResponse']
_GETTRACKERTABLEREQUEST = DESCRIPTOR.message_types_by_name['GetTrackerTableRequest']
_TRACKERTABLE = DESCRIPTOR.message_types_by_name['TrackerTable']
_TRACKERDATASETREQUEST = DESCRIPTOR.message_types_by_name['TrackerDatasetRequest']
_TRACKERDATASETREQUEST_LOGIDS = _TRACKERDATASETREQUEST.nested_types_by_name['LogIds']
_TRACKERDATASET = DESCRIPTOR.message_types_by_name['TrackerDataset']
_RECORDCOLUMN_DATATYPE = _RECORDCOLUMN.enum_types_by_name['DataType']
_TRACKERDATASETREQUEST_EXPORTFORMAT = _TRACKERDATASETREQUEST.enum_types_by_name['ExportFormat']
TrackerLog = _reflection.GeneratedProtocolMessageType('TrackerLog', (_message.Message,), {
  'DESCRIPTOR' : _TRACKERLOG,
  '__module__' : 'logs_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.TrackerLog)
  })
_sym_db.RegisterMessage(TrackerLog)

TrackerLogId = _reflection.GeneratedProtocolMessageType('TrackerLogId', (_message.Message,), {
  'DESCRIPTOR' : _TRACKERLOGID,
  '__module__' : 'logs_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.TrackerLogId)
  })
_sym_db.RegisterMessage(TrackerLogId)

RecordColumn = _reflection.GeneratedProtocolMessageType('RecordColumn', (_message.Message,), {

  'RecordItem' : _reflection.GeneratedProtocolMessageType('RecordItem', (_message.Message,), {
    'DESCRIPTOR' : _RECORDCOLUMN_RECORDITEM,
    '__module__' : 'logs_pb2'
    # @@protoc_insertion_point(class_scope:lmao_pb.RecordColumn.RecordItem)
    })
  ,
  'DESCRIPTOR' : _RECORDCOLUMN,
  '__module__' : 'logs_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.RecordColumn)
  })
_sym_db.RegisterMessage(RecordColumn)
_sym_db.RegisterMessage(RecordColumn.RecordItem)

RecordRow = _reflection.GeneratedProtocolMessageType('RecordRow', (_message.Message,), {
  'DESCRIPTOR' : _RECORDROW,
  '__module__' : 'logs_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.RecordRow)
  })
_sym_db.RegisterMessage(RecordRow)

TrackerLogRequest = _reflection.GeneratedProtocolMessageType('TrackerLogRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRACKERLOGREQUEST,
  '__module__' : 'logs_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.TrackerLogRequest)
  })
_sym_db.RegisterMessage(TrackerLogRequest)

TrackerLogResponse = _reflection.GeneratedProtocolMessageType('TrackerLogResponse', (_message.Message,), {
  'DESCRIPTOR' : _TRACKERLOGRESPONSE,
  '__module__' : 'logs_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.TrackerLogResponse)
  })
_sym_db.RegisterMessage(TrackerLogResponse)

GetTrackerTableRequest = _reflection.GeneratedProtocolMessageType('GetTrackerTableRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTRACKERTABLEREQUEST,
  '__module__' : 'logs_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.GetTrackerTableRequest)
  })
_sym_db.RegisterMessage(GetTrackerTableRequest)

TrackerTable = _reflection.GeneratedProtocolMessageType('TrackerTable', (_message.Message,), {
  'DESCRIPTOR' : _TRACKERTABLE,
  '__module__' : 'logs_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.TrackerTable)
  })
_sym_db.RegisterMessage(TrackerTable)

TrackerDatasetRequest = _reflection.GeneratedProtocolMessageType('TrackerDatasetRequest', (_message.Message,), {

  'LogIds' : _reflection.GeneratedProtocolMessageType('LogIds', (_message.Message,), {
    'DESCRIPTOR' : _TRACKERDATASETREQUEST_LOGIDS,
    '__module__' : 'logs_pb2'
    # @@protoc_insertion_point(class_scope:lmao_pb.TrackerDatasetRequest.LogIds)
    })
  ,
  'DESCRIPTOR' : _TRACKERDATASETREQUEST,
  '__module__' : 'logs_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.TrackerDatasetRequest)
  })
_sym_db.RegisterMessage(TrackerDatasetRequest)
_sym_db.RegisterMessage(TrackerDatasetRequest.LogIds)

TrackerDataset = _reflection.GeneratedProtocolMessageType('TrackerDataset', (_message.Message,), {
  'DESCRIPTOR' : _TRACKERDATASET,
  '__module__' : 'logs_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.TrackerDataset)
  })
_sym_db.RegisterMessage(TrackerDataset)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z-github.com/NimbleBoxAI/nimblebox-lmao/lmao_pb'
  _TRACKERLOG._serialized_start=134
  _TRACKERLOG._serialized_end=355
  _TRACKERLOGID._serialized_start=357
  _TRACKERLOGID._serialized_end=387
  _RECORDCOLUMN._serialized_start=390
  _RECORDCOLUMN._serialized_end=682
  _RECORDCOLUMN_RECORDITEM._serialized_start=519
  _RECORDCOLUMN_RECORDITEM._serialized_end=635
  _RECORDCOLUMN_DATATYPE._serialized_start=637
  _RECORDCOLUMN_DATATYPE._serialized_end=682
  _RECORDROW._serialized_start=684
  _RECORDROW._serialized_end=755
  _TRACKERLOGREQUEST._serialized_start=758
  _TRACKERLOGREQUEST._serialized_end=972
  _TRACKERLOGRESPONSE._serialized_start=974
  _TRACKERLOGRESPONSE._serialized_end=1031
  _GETTRACKERTABLEREQUEST._serialized_start=1034
  _GETTRACKERTABLEREQUEST._serialized_end=1175
  _TRACKERTABLE._serialized_start=1177
  _TRACKERTABLE._serialized_end=1232
  _TRACKERDATASETREQUEST._serialized_start=1235
  _TRACKERDATASETREQUEST._serialized_end=1640
  _TRACKERDATASETREQUEST_LOGIDS._serialized_start=1559
  _TRACKERDATASETREQUEST_LOGIDS._serialized_end=1584
  _TRACKERDATASETREQUEST_EXPORTFORMAT._serialized_start=1586
  _TRACKERDATASETREQUEST_EXPORTFORMAT._serialized_end=1625
  _TRACKERDATASET._serialized_start=1642
  _TRACKERDATASET._serialized_end=1725
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tracker.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rtracker.proto\x12\x07lmao_pb\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xde\x01\n\x12InitTrackerRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x1a\n\x04type\x18\x03 \x01(\x0e\x32\x0c.lmao_pb.NBX\x12\x14\n\x0cnbx_group_id\x18\x04 \x01(\t\x12\x17\n\x0fnbx_instance_id\x18\x05 \x01(\t\x12\'\n\x06\x63onfig\x18\x06 \x01(\x0b\x32\x17.google.protobuf.Struct\x12*\n\x0ctracker_type\x18\x07 \x01(\x0e\x32\x14.lmao_pb.TrackerType\"\xd8\x03\n\x07Tracker\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\x12\'\n\x06\x63onfig\x18\x05 \x01(\x0b\x32\x17.google.protobuf.Struct\x12.\n\ncreated_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nupdated_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x15\n\rsave_location\x18\t \x01(\t\x12\'\n\x06status\x18\n \x01(\x0e\x32\x17.lmao_pb.Tracker.Status\x12*\n\x0ctracker_type\x18\x0b \x01(\x0e\x32\x14.lmao_pb.TrackerType\x12\x1a\n\x04type\x18\x0c \x01(\x0e\x32\x0c.lmao_pb.NBX\x12\x14\n\x0cnbx_group_id\x18\r \x01(\t\x12\x17\n\x0fnbx_instance_id\x18\x0e \x01(\t\x12\x13\n\x0bupdate_keys\x18\x0f \x03(\t\"B\n\x06Status\x12\x10\n\x0cUNSET_STATUS\x10\x00\x12\x0b\n\x07RUNNING\x10\x01\x12\r\n\tCOMPLETED\x10\x02\x12\n\n\x06\x46\x41ILED\x10\x03\"\x81\x03\n\x13ListTrackersRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x0e\n\x06offset\x18\x03 \x01(\x03\x12\r\n\x05limit\x18\x04 \x01(\x03\x12\x31\n\rcreated_after\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x32\n\x0e\x63reated_before\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\rupdated_after\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x32\n\x0eupdated_before\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x06status\x18\t \x01(\x0e\x32\x17.lmao_pb.Tracker.Status\x12*\n\x0ctracker_type\x18\n \x01(\x0e\x32\x14.lmao_pb.TrackerType\"O\n\x14ListTrackersResponse\x12\"\n\x08trackers\x18\x01 \x03(\x0b\x32\x10.lmao_pb.Tracker\x12\x13\n\x0btotal_pages\x18\x02 \x01(\x05*&\n\x03NBX\x12\t\n\x05LOCAL\x10\x00\x12\x07\n\x03JOB\x10\x01\x12\x0b\n\x07SERVING\x10\x02*7\n\x0bTrackerType\x12\x0e\n\nUNSET_TYPE\x10\x00\x12\x0e\n\nEXPERIMENT\x10\x01\x12\x08\n\x04LIVE\x10\x02\x42/Z-github.com/NimbleBoxAI/nimblebox-lmao/lmao_pbb\x06proto3')

_NBX = DESCRIPTOR.enum_types_by_name['NBX']
NBX = enum_type_wrapper.EnumTypeWrapper(_NBX)
_TRACKERTYPE = DESCRIPTOR.enum_types_by_name['TrackerType']
TrackerType = enum_type_wrapper.EnumTypeWrapper(_TRACKERTYPE)
LOCAL = 0
JOB = 1
SERVING = 2
UNSET_TYPE = 0
EXPERIMENT = 1
LIVE = 2


_INITTRACKERREQUEST = DESCRIPTOR.message_types_by_name['InitTrackerRequest']
_TRACKER = DESCRIPTOR.message_types_by_name['Tracker']
_LISTTRACKERSREQUEST = DESCRIPTOR.message_types_by_name['ListTrackersRequest']
_LISTTRACKERSRESPONSE = DESCRIPTOR.message_types_by_name['ListTrackersResponse']
_TRACKER_STATUS = _TRACKER.enum_types_by_name['Status']
InitTrackerRequest = _reflection.GeneratedProtocolMessageType('InitTrackerRequest', (_message.Message,), {
  'DESCRIPTOR' : _INITTRACKERREQUEST,
  '__module__' : 'tracker_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.InitTrackerRequest)
  })
_sym_db.RegisterMessage(InitTrackerRequest)

Tracker = _reflection.GeneratedProtocolMessageType('Tracker', (_message.Message,), {
  'DESCRIPTOR' : _TRACKER,
  '__module__' : 'tracker_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.Tracker)
  })
_sym_db.RegisterMessage(Tracker)

ListTrackersRequest = _reflection.GeneratedProtocolMessageType('ListTrackersRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTTRACKERSREQUEST,
  '__module__' : 'tracker_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.ListTrackersRequest)
  })
_sym_db.RegisterMessage(ListTrackersRequest)

ListTrackersResponse = _reflection.GeneratedProtocolMessageType('ListTrackersResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTTRACKERSRESPONSE,
  '__module__' : 'tracker_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.ListTrackersResponse)
  })
_sym_db.RegisterMessage(ListTrackersResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z-github.com/NimbleBoxAI/nimblebox-lmao/lmao_pb'
  _NBX._serialized_start=1258
  _NBX._serialized_end=1296
  _TRACKERTYPE._serialized_start=1298
  _TRACKERTYPE._serialized_end=1353
  _INITTRACKERREQUEST._serialized_start=90
  _INITTRACKERREQUEST._serialized_end=312
  _TRACKER._serialized_start=315
  _TRACKER._serialized_end=787
  _TRACKER_STATUS._serialized_start=721
  _TRACKER_STATUS._serialized_end=787
  _LISTTRACKERSREQUEST._serialized_start=790
  _LISTTRACKERSREQUEST._serialized_end=1175
  _LISTTRACKERSRESPONSE._serialized_start=1177
  _LISTTRACKERSRESPONSE._serialized_end=1256
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.struct_pb2
import google.protobuf.timestamp_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Project(google.protobuf.message.Message):
    """
    Project level messages

    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    TOTAL_RUNS_FIELD_NUMBER: builtins.int
    TOTAL_SERVINGS_FIELD_NUMBER: builtins.int
    CREATED_AT_FIELD_NUMBER: builtins.int
    UPDATED_AT_FIELD_NUMBER: builtins.int
    RELIC_ID_FIELD_NUMBER: builtins.int
    UPDATE_KEYS_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    RUNS_COUNTER_FIELD_NUMBER: builtins.int
    LIVE_COUNTER_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the workspace this is part of"""

    id: typing.Text
    """the unique ID for this project"""

    name: typing.Text
    """the usermodifiable name of the project"""

    total_runs: builtins.int
    """the total number of experiments in this project"""

    total_servings: builtins.int
    """the total number of live deployments in this project"""

    @property
    def created_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """the time when this project was created"""
        pass
    @property
    def updated_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """the time when this project was created"""
        pass
    relic_id: typing.Text
    """the relic ID for this project"""

    @property
    def update_keys(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """the list of update keys for this project"""
        pass
    @property
    def metadata(self) -> google.protobuf.struct_pb2.Struct:
        """the metadata for this project"""
        pass
    runs_counter: builtins.int
    """currently active runs"""

    live_counter: builtins.int
    """currently active live deployments"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        id: typing.Text = ...,
        name: typing.Text = ...,
        total_runs: builtins.int = ...,
        total_servings: builtins.int = ...,
        created_at: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        updated_at: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        relic_id: typing.Text = ...,
        update_keys: typing.Optional[typing.Iterable[typing.Text]] = ...,
        metadata: typing.Optional[google.protobuf.struct_pb2.Struct] = ...,
        runs_counter: builtins.int = ...,
        live_counter: builtins.int = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["created_at",b"created_at","metadata",b"metadata","updated_at",b"updated_at"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["created_at",b"created_at","id",b"id","live_counter",b"live_counter","metadata",b"metadata","name",b"name","relic_id",b"relic_id","runs_counter",b"runs_counter","total_runs",b"total_runs","total_servings",b"total_servings","update_keys",b"update_keys","updated_at",b"updated_at","workspace_id",b"workspace_id"]) -> None: ...
global___Project = Project

class InitProjectRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the workspace this LMAO project is part of"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["workspace_id",b"workspace_id"]) -> None: ...
global___InitProjectRequest = InitProjectRequest

class InitProjectResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    RELIC_ID_FIELD_NUMBER: builtins.int
    RELIC_NAME_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    id: typing.Text
    relic_id: typing.Text
    relic_name: typing.Text
    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        id: typing.Text = ...,
        relic_id: typing.Text = ...,
        relic_name: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["id",b"id","relic_id",b"relic_id","relic_name",b"relic_name","workspace_id",b"workspace_id"]) -> None: ...
global___InitProjectResponse = InitProjectResponse

class ListProjectsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    PROJECT_ID_OR_NAME_FIELD_NUMBER: builtins.int
    OFFSET_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    CREATED_AFTER_FIELD_NUMBER: builtins.int
    CREATED_BEFORE_FIELD_NUMBER: builtins.int
    UPDATED_AFTER_FIELD_NUMBER: builtins.int
    UPDATED_BEFORE_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the workspace this is part of"""

    project_id_or_name: typing.Text
    """used specifically for searching, if empty, it will return all the projects"""

    offset: builtins.int
    """the offset for pagination"""

    limit: builtins.int
    """the limit for pagination"""

    @property
    def created_after(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """the time after which the project was created"""
        pass
    @property
    def created_before(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """the time before which the project was created"""
        pass
    @property
    def updated_after(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """the time after which the project was updated"""
        pass
    @property
    def updated_before(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """the time before which the project was updated"""
        pass
    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        project_id_or_name: typing.Text = ...,
        offset: builtins.int = ...,
        limit: builtins.int = ...,
        created_after: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        created_before: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        updated_after: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        updated_before: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["created_after",b"created_after","created_before",b"created_before","updated_after",b"updated_after","updated_before",b"updated_before"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["created_after",b"created_after","created_before",b"created_before","limit",b"limit","offset",b"offset","project_id_or_name",b"project_id_or_name","updated_after",b"updated_after","updated_before",b"updated_before","workspace_id",b"workspace_id"]) -> None: ...
global___ListProjectsRequest = ListProjectsRequest

class ListProjectsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    PROJECTS_FIELD_NUMBER: builtins.int
    @property
    def projects(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Project]:
        """the list of projects in this workspace"""
        pass
    def __init__(self,
        *,
        projects: typing.Optional[typing.Iterable[global___Project]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["projects",b"projects"]) -> None: ...
global___ListProjectsResponse = ListProjectsResponse



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: project.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rproject.proto\x12\x07lmao_pb\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xc3\x02\n\x07Project\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x12\n\ntotal_runs\x18\x04 \x01(\x03\x12\x16\n\x0etotal_servings\x18\x05 \x01(\x03\x12.\n\ncreated_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nupdated_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08relic_id\x18\x08 \x01(\t\x12\x13\n\x0bupdate_keys\x18\t \x03(\t\x12)\n\x08metadata\x18\n \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x14\n\x0cruns_counter\x18\x0b \x01(\x05\x12\x14\n\x0clive_counter\x18\x0c \x01(\x05\"*\n\x12InitProjectRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\"]\n\x13InitProjectResponse\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12\x10\n\x08relic_id\x18\x03 \x01(\t\x12\x12\n\nrelic_name\x18\x04 \x01(\t\"\xb4\x02\n\x13ListProjectsRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x1a\n\x12project_id_or_name\x18\x02 \x01(\t\x12\x0e\n\x06offset\x18\x03 \x01(\x03\x12\r\n\x05limit\x18\x04 \x01(\x03\x12\x31\n\rcreated_after\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x32\n\x0e\x63reated_before\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\rupdated_after\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x32\n\x0eupdated_before\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\":\n\x14ListProjectsResponse\x12\"\n\x08projects\x18\x01 \x03(\x0b\x32\x10.lmao_pb.ProjectB/Z-github.com/NimbleBoxAI/nimblebox-lmao/lmao_pbb\x06proto3')



_PROJECT = DESCRIPTOR.message_types_by_name['Project']
_INITPROJECTREQUEST = DESCRIPTOR.message_types_by_name['InitProjectRequest']
_INITPROJECTRESPONSE = DESCRIPTOR.message_types_by_name['InitProjectResponse']
_LISTPROJECTSREQUEST = DESCRIPTOR.message_types_by_name['ListProjectsRequest']
_LISTPROJECTSRESPONSE = DESCRIPTOR.message_types_by_name['ListProjectsResponse']
Project = _reflection.GeneratedProtocolMessageType('Project', (_message.Message,), {
  'DESCRIPTOR' : _PROJECT,
  '__module__' : 'project_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.Project)
  })
_sym_db.RegisterMessage(Project)

InitProjectRequest = _reflection.GeneratedProtocolMessageType('InitProjectRequest', (_message.Message,), {
  'DESCRIPTOR' : _INITPROJECTREQUEST,
  '__module__' : 'project_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.InitProjectRequest)
  })
_sym_db.RegisterMessage(InitProjectRequest)

InitProjectResponse = _reflection.GeneratedProtocolMessageType('InitProjectResponse', (_message.Message,), {
  'DESCRIPTOR' : _INITPROJECTRESPONSE,
  '__module__' : 'project_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.InitProjectResponse)
  })
_sym_db.RegisterMessage(InitProjectResponse)

ListProjectsRequest = _reflection.GeneratedProtocolMessageType('ListProjectsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPROJECTSREQUEST,
  '__module__' : 'project_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.ListProjectsRequest)
  })
_sym_db.RegisterMessage(ListProjectsRequest)

ListProjectsResponse = _reflection.GeneratedProtocolMessageType('ListProjectsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPROJECTSRESPONSE,
  '__module__' : 'project_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.ListProjectsResponse)
  })
_sym_db.RegisterMessage(ListProjectsResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z-github.com/NimbleBoxAI/nimblebox-lmao/lmao_pb'
  _PROJECT._serialized_start=90
  _PROJECT._serialized_end=413
  _INITPROJECTREQUEST._serialized_start=415
  _INITPROJECTREQUEST._serialized_end=457
  _INITPROJECTRESPONSE._serialized_start=459
  _INITPROJECTRESPONSE._serialized_end=552
  _LISTPROJECTSREQUEST._serialized_start=555
  _LISTPROJECTSREQUEST._serialized_end=863
  _LISTPROJECTSRESPONSE._serialized_start=865
  _LISTPROJECTSRESPONSE._serialized_end=923
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.struct_pb2
import google.protobuf.timestamp_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _NBX:
    ValueType = typing.NewType('ValueType', builtins.int)
    V: typing_extensions.TypeAlias = ValueType
class _NBXEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_NBX.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    LOCAL: _NBX.ValueType  # 0
    JOB: _NBX.ValueType  # 1
    SERVING: _NBX.ValueType  # 2
class NBX(_NBX, metaclass=_NBXEnumTypeWrapper):
    """
    Trackers are the new architecture we using, thing like this is a singular client that can
    be used by anyone to track metrics, logs, etc.

    """
    pass

LOCAL: NBX.ValueType  # 0
JOB: NBX.ValueType  # 1
SERVING: NBX.ValueType  # 2
global___NBX = NBX


class _TrackerType:
    ValueType = typing.NewType('ValueType', builtins.int)
    V: typing_extensions.TypeAlias = ValueType
class _TrackerTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_TrackerType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNSET_TYPE: _TrackerType.ValueType  # 0
    """this is not allowed in the DB, but given here for completeness"""

    EXPERIMENT: _TrackerType.ValueType  # 1
    """this is a tracker for an experiment"""

    LIVE: _TrackerType.ValueType  # 2
    """this is a tracker for a live run"""

class TrackerType(_TrackerType, metaclass=_TrackerTypeEnumTypeWrapper):
    pass

UNSET_TYPE: TrackerType.ValueType  # 0
"""this is not allowed in the DB, but given here for completeness"""

EXPERIMENT: TrackerType.ValueType  # 1
"""this is a tracker for an experiment"""

LIVE: TrackerType.ValueType  # 2
"""this is a tracker for a live run"""

global___TrackerType = TrackerType


class InitTrackerRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    PROJECT_ID_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    NBX_GROUP_ID_FIELD_NUMBER: builtins.int
    NBX_INSTANCE_ID_FIELD_NUMBER: builtins.int
    CONFIG_FIELD_NUMBER: builtins.int
    TRACKER_TYPE_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    project_id: typing.Text
    type: global___NBX.ValueType
    """this is the type of the agent, JOB or SERVING"""

    nbx_group_id: typing.Text
    """this is the group ID of the NBX-Job or Serving"""

    nbx_instance_id: typing.Text
    """this is the instance ID of the NBX-Job or Serving"""

    @property
    def config(self) -> google.protobuf.struct_pb2.Struct:
        """anything else that the user wants to log"""
        pass
    tracker_type: global___TrackerType.ValueType
    """the type of the tracker"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        project_id: typing.Text = ...,
        type: global___NBX.ValueType = ...,
        nbx_group_id: typing.Text = ...,
        nbx_instance_id: typing.Text = ...,
        config: typing.Optional[google.protobuf.struct_pb2.Struct] = ...,
        tracker_type: global___TrackerType.ValueType = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["config",b"config"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["config",b"config","nbx_group_id",b"nbx_group_id","nbx_instance_id",b"nbx_instance_id","project_id",b"project_id","tracker_type",b"tracker_type","type",b"type","workspace_id",b"workspace_id"]) -> None: ...
global___InitTrackerRequest = InitTrackerRequest

class Tracker(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _Status:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _StatusEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Tracker._Status.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSET_STATUS: Tracker._Status.ValueType  # 0
        """When created and added to DB but no further action taken"""

        RUNNING: Tracker._Status.ValueType  # 1
        """When the first on_log has been called"""

        COMPLETED: Tracker._Status.ValueType  # 2
        """When the tracking has ended"""

        FAILED: Tracker._Status.ValueType  # 3
        """When the NBX-Run/Model failed"""

    class Status(_Status, metaclass=_StatusEnumTypeWrapper):
        pass

    UNSET_STATUS: Tracker.Status.ValueType  # 0
    """When created and added to DB but no further action taken"""

    RUNNING: Tracker.Status.ValueType  # 1
    """When the first on_log has been called"""

    COMPLETED: Tracker.Status.ValueType  # 2
    """When the tracking has ended"""

    FAILED: Tracker.Status.ValueType  # 3
    """When the NBX-Run/Model failed"""


    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    PROJECT_ID_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    CONFIG_FIELD_NUMBER: builtins.int
    CREATED_AT_FIELD_NUMBER: builtins.int
    UPDATED_AT_FIELD_NUMBER: builtins.int
    SAVE_LOCATION_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    TRACKER_TYPE_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    NBX_GROUP_ID_FIELD_NUMBER: builtins.int
    NBX_INSTANCE_ID_FIELD_NUMBER: builtins.int
    UPDATE_KEYS_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    project_id: typing.Text
    id: typing.Text
    """this is the unique tracker ID"""

    @property
    def config(self) -> google.protobuf.struct_pb2.Struct:
        """anything else that the user wants to log"""
        pass
    @property
    def created_at(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    @property
    def updated_at(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    save_location: typing.Text
    """the location where the files are saved, this will be pulled by the server"""

    status: global___Tracker.Status.ValueType
    """the last known status of this tracker"""

    tracker_type: global___TrackerType.ValueType
    """the type of the tracker"""

    type: global___NBX.ValueType
    """this is the type of the agent, JOB or SERVING"""

    nbx_group_id: typing.Text
    """this is the group ID of the NBX-Job or Serving"""

    nbx_instance_id: typing.Text
    """this is the instance ID of the NBX-Job or Serving"""

    @property
    def update_keys(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """now we create a special thing which will be populated and sent during updates"""
        pass
    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        project_id: typing.Text = ...,
        id: typing.Text = ...,
        config: typing.Optional[google.protobuf.struct_pb2.Struct] = ...,
        created_at: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        updated_at: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        save_location: typing.Text = ...,
        status: global___Tracker.Status.ValueType = ...,
        tracker_type: global___TrackerType.ValueType = ...,
        type: global___NBX.ValueType = ...,
        nbx_group_id: typing.Text = ...,
        nbx_instance_id: typing.Text = ...,
        update_keys: typing.Optional[typing.Iterable[typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["config",b"config","created_at",b"created_at","updated_at",b"updated_at"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["config",b"config","created_at",b"created_at","id",b"id","nbx_group_id",b"nbx_group_id","nbx_instance_id",b"nbx_instance_id","project_id",b"project_id","save_location",b"save_location","status",b"status","tracker_type",b"tracker_type","type",b"type","update_keys",b"update_keys","updated_at",b"updated_at","workspace_id",b"workspace_id"]) -> None: ...
global___Tracker = Tracker

class ListTrackersRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    PROJECT_ID_FIELD_NUMBER: builtins.int
    OFFSET_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    CREATED_AFTER_FIELD_NUMBER: builtins.int
    CREATED_BEFORE_FIELD_NUMBER: builtins.int
    UPDATED_AFTER_FIELD_NUMBER: builtins.int
    UPDATED_BEFORE_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    TRACKER_TYPE_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    project_id: typing.Text
    offset: builtins.int
    """the offset for pagination"""

    limit: builtins.int
    """the limit for pagination"""

    @property
    def created_after(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """the time after which the project was created"""
        pass
    @property
    def created_before(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """the time before which the project was created"""
        pass
    @property
    def updated_after(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """the time after which the project was updated"""
        pass
    @property
    def updated_before(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """the time before which the project was updated"""
        pass
    status: global___Tracker.Status.ValueType
    """the status of the tracker"""

    tracker_type: global___TrackerType.ValueType
    """the type of the tracker"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        project_id: typing.Text = ...,
        offset: builtins.int = ...,
        limit: builtins.int = ...,
        created_after: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        created_before: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        updated_after: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        updated_before: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        status: global___Tracker.Status.ValueType = ...,
        tracker_type: global___TrackerType.ValueType = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["created_after",b"created_after","created_before",b"created_before","updated_after",b"updated_after","updated_before",b"updated_before"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["created_after",b"created_after","created_before",b"created_before","limit",b"limit","offset",b"offset","project_id",b"project_id","status",b"status","tracker_type",b"tracker_type","updated_after",b"updated_after","updated_before",b"updated_before","workspace_id",b"workspace_id"]) -> None: ...
global___ListTrackersRequest = ListTrackersRequest

class ListTrackersResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    TRACKERS_FIELD_NUMBER: builtins.int
    TOTAL_PAGES_FIELD_NUMBER: builtins.int
    @property
    def trackers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Tracker]: ...
    total_pages: builtins.int
    """the total number of pages"""

    def __init__(self,
        *,
        trackers: typing.Optional[typing.Iterable[global___Tracker]] = ...,
        total_pages: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["total_pages",b"total_pages","trackers",b"trackers"]) -> None: ...
global___ListTrackersResponse = ListTrackersResponse



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import google.protobuf.empty_pb2
import grpc
try:
  from gen.proto import logs_pb2 as logs__pb2
except ImportError:
  from nbox.lmao_v4.proto import logs_pb2
try:
  from gen.proto import project_pb2 as project__pb2
except ImportError:
  from nbox.lmao_v4.proto import project_pb2
try:
  from gen.proto import rules_pb2 as rules__pb2
except ImportError:
  from nbox.lmao_v4.proto import rules_pb2
try:
  from gen.proto import tracker_pb2 as tracker__pb2
except ImportError:
  from nbox.lmao_v4.proto import tracker_pb2

class LMAOStub:
    """This is the service definition for the LMAO server, this server will talk to the MongoDB in the backend."""
    def __init__(self, channel: grpc.Channel) -> None: ...
    InitProject: grpc.UnaryUnaryMultiCallable[
        project_pb2.InitProjectRequest,
        project_pb2.InitProjectResponse]
    """Project level information -> Full CRUDL on project"""

    GetProject: grpc.UnaryUnaryMultiCallable[
        project_pb2.Project,
        project_pb2.Project]

    UpdateProject: grpc.UnaryUnaryMultiCallable[
        project_pb2.Project,
        project_pb2.Project]

    DeleteProject: grpc.UnaryUnaryMultiCallable[
        project_pb2.Project,
        google.protobuf.empty_pb2.Empty]

    ListProjects: grpc.UnaryUnaryMultiCallable[
        project_pb2.ListProjectsRequest,
        project_pb2.ListProjectsResponse]

    GetRuleBuilder: grpc.UnaryUnaryMultiCallable[
        rules_pb2.RuleBuilder,
        rules_pb2.RuleBuilder]
    """Rules based workflow -> Full CRUDL for rules"""

    CreateRule: grpc.UnaryUnaryMultiCallable[
        rules_pb2.InitRuleRequest,
        rules_pb2.Rule]

    GetRule: grpc.UnaryUnaryMultiCallable[
        rules_pb2.Rule,
        rules_pb2.Rule]

    UpdateRule: grpc.UnaryUnaryMultiCallable[
        rules_pb2.Rule,
        rules_pb2.Rule]

    DeleteRule: grpc.UnaryUnaryMultiCallable[
        rules_pb2.Rule,
        google.protobuf.empty_pb2.Empty]

    ListRules: grpc.UnaryUnaryMultiCallable[
        rules_pb2.RulesList,
        rules_pb2.RulesList]

    InitTracker: grpc.UnaryUnaryMultiCallable[
        tracker_pb2.InitTrackerRequest,
        tracker_pb2.Tracker]
    """Generic Tracking -> Full CRUDL on trackers"""

    GetTracker: grpc.UnaryUnaryMultiCallable[
        tracker_pb2.Tracker,
        tracker_pb2.Tracker]

    UpdateTracker: grpc.UnaryUnaryMultiCallable[
        tracker_pb2.Tracker,
        tracker_pb2.Tracker]

    DeleteTracker: grpc.UnaryUnaryMultiCallable[
        tracker_pb2.Tracker,
        google.protobuf.empty_pb2.Empty]

    ListTrackers: grpc.UnaryUnaryMultiCallable[
        tracker_pb2.ListTrackersRequest,
        tracker_pb2.ListTrackersResponse]

    GetTrackerTable: grpc.UnaryUnaryMultiCallable[
        logs_pb2.GetTrackerTableRequest,
        logs_pb2.TrackerTable]
    """things for logs"""

    PutTrackerLog: grpc.UnaryUnaryMultiCallable[
        logs_pb2.TrackerLog,
        logs_pb2.TrackerLogId]

    GetTrackerLogs: grpc.UnaryUnaryMultiCallable[
        logs_pb2.TrackerLogRequest,
        logs_pb2.TrackerLogResponse]

    CreateDataset: grpc.UnaryUnaryMultiCallable[
        logs_pb2.TrackerDatasetRequest,
        logs_pb2.TrackerDataset]
    """
    This is the new service that we want to add

    """

    CreateDatasetStatus: grpc.UnaryUnaryMultiCallable[
        logs_pb2.TrackerDataset,
        logs_pb2.TrackerDataset]


class LMAOServicer(metaclass=abc.ABCMeta):
    """This is the service definition for the LMAO server, this server will talk to the MongoDB in the backend."""
    @abc.abstractmethod
    def InitProject(self,
        request: project_pb2.InitProjectRequest,
        context: grpc.ServicerContext,
    ) -> project_pb2.InitProjectResponse:
        """Project level information -> Full CRUDL on project"""
        pass

    @abc.abstractmethod
    def GetProject(self,
        request: project_pb2.Project,
        context: grpc.ServicerContext,
    ) -> project_pb2.Project: ...

    @abc.abstractmethod
    def UpdateProject(self,
        request: project_pb2.Project,
        context: grpc.ServicerContext,
    ) -> project_pb2.Project: ...

    @abc.abstractmethod
    def DeleteProject(self,
        request: project_pb2.Project,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...

    @abc.abstractmethod
    def ListProjects(self,
        request: project_pb2.ListProjectsRequest,
        context: grpc.ServicerContext,
    ) -> project_pb2.ListProjectsResponse: ...

    @abc.abstractmethod
    def GetRuleBuilder(self,
        request: rules_pb2.RuleBuilder,
        context: grpc.ServicerContext,
    ) -> rules_pb2.RuleBuilder:
        """Rules based workflow -> Full CRUDL for rules"""
        pass

    @abc.abstractmethod
    def CreateRule(self,
        request: rules_pb2.InitRuleRequest,
        context: grpc.ServicerContext,
    ) -> rules_pb2.Rule: ...

    @abc.abstractmethod
    def GetRule(self,
        request: rules_pb2.Rule,
        context: grpc.ServicerContext,
    ) -> rules_pb2.Rule: ...

    @abc.abstractmethod
    def UpdateRule(self,
        request: rules_pb2.Rule,
        context: grpc.ServicerContext,
    ) -> rules_pb2.Rule: ...

    @abc.abstractmethod
    def DeleteRule(self,
        request: rules_pb2.Rule,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...

    @abc.abstractmethod
    def ListRules(self,
        request: rules_pb2.RulesList,
        context: grpc.ServicerContext,
    ) -> rules_pb2.RulesList: ...

    @abc.abstractmethod
    def InitTracker(self,
        request: tracker_pb2.InitTrackerRequest,
        context: grpc.ServicerContext,
    ) -> tracker_pb2.Tracker:
        """Generic Tracking -> Full CRUDL on trackers"""
        pass

    @abc.abstractmethod
    def GetTracker(self,
        request: tracker_pb2.Tracker,
        context: grpc.ServicerContext,
    ) -> tracker_pb2.Tracker: ...

    @abc.abstractmethod
    def UpdateTracker(self,
        request: tracker_pb2.Tracker,
        context: grpc.ServicerContext,
    ) -> tracker_pb2.Tracker: ...

    @abc.abstractmethod
    def DeleteTracker(self,
        request: tracker_pb2.Tracker,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty: ...

    @abc.abstractmethod
    def ListTrackers(self,
        request: tracker_pb2.ListTrackersRequest,
        context: grpc.ServicerContext,
    ) -> tracker_pb2.ListTrackersResponse: ...

    @abc.abstractmethod
    def GetTrackerTable(self,
        request: logs_pb2.GetTrackerTableRequest,
        context: grpc.ServicerContext,
    ) -> logs_pb2.TrackerTable:
        """things for logs"""
        pass

    @abc.abstractmethod
    def PutTrackerLog(self,
        request: logs_pb2.TrackerLog,
        context: grpc.ServicerContext,
    ) -> logs_pb2.TrackerLogId: ...

    @abc.abstractmethod
    def GetTrackerLogs(self,
        request: logs_pb2.TrackerLogRequest,
        context: grpc.ServicerContext,
    ) -> logs_pb2.TrackerLogResponse: ...

    @abc.abstractmethod
    def CreateDataset(self,
        request: logs_pb2.TrackerDatasetRequest,
        context: grpc.ServicerContext,
    ) -> logs_pb2.TrackerDataset:
        """
        This is the new service that we want to add

        """
        pass

    @abc.abstractmethod
    def CreateDatasetStatus(self,
        request: logs_pb2.TrackerDataset,
        context: grpc.ServicerContext,
    ) -> logs_pb2.TrackerDataset: ...


def add_LMAOServicer_to_server(servicer: LMAOServicer, server: grpc.Server) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rules.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0brules.proto\x12\x07lmao_pb\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\x9f\x01\n\x05Logic\x12\x13\n\x0b\x61\x63tion_type\x18\x01 \x01(\t\x12,\n\x0b\x61\x63tion_data\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12&\n\x05\x65xtra\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\x12+\n\njson_logic\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\"m\n\x0fInitRuleRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x11\n\trule_name\x18\x04 \x01(\t\x12\x1d\n\x05logic\x18\x07 \x01(\x0b\x32\x0e.lmao_pb.Logic\"\xfb\x01\n\x04Rule\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\x12\x11\n\trule_name\x18\x04 \x01(\t\x12.\n\ncreated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nupdated_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x16\n\x0eis_deactivated\x18\x07 \x01(\x08\x12\x1d\n\x05logic\x18\x08 \x01(\x0b\x32\x0e.lmao_pb.Logic\x12\x13\n\x0bupdate_keys\x18\t \x03(\t\"S\n\tRulesList\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x1c\n\x05rules\x18\x03 \x03(\x0b\x32\r.lmao_pb.Rule\"V\n\x0bRuleBuilder\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x1d\n\x05logic\x18\x03 \x01(\x0b\x32\x0e.lmao_pb.LogicB/Z-github.com/NimbleBoxAI/nimblebox-lmao/lmao_pbb\x06proto3')



_LOGIC = DESCRIPTOR.message_types_by_name['Logic']
_INITRULEREQUEST = DESCRIPTOR.message_types_by_name['InitRuleRequest']
_RULE = DESCRIPTOR.message_types_by_name['Rule']
_RULESLIST = DESCRIPTOR.message_types_by_name['RulesList']
_RULEBUILDER = DESCRIPTOR.message_types_by_name['RuleBuilder']
Logic = _reflection.GeneratedProtocolMessageType('Logic', (_message.Message,), {
  'DESCRIPTOR' : _LOGIC,
  '__module__' : 'rules_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.Logic)
  })
_sym_db.RegisterMessage(Logic)

InitRuleRequest = _reflection.GeneratedProtocolMessageType('InitRuleRequest', (_message.Message,), {
  'DESCRIPTOR' : _INITRULEREQUEST,
  '__module__' : 'rules_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.InitRuleRequest)
  })
_sym_db.RegisterMessage(InitRuleRequest)

Rule = _reflection.GeneratedProtocolMessageType('Rule', (_message.Message,), {
  'DESCRIPTOR' : _RULE,
  '__module__' : 'rules_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.Rule)
  })
_sym_db.RegisterMessage(Rule)

RulesList = _reflection.GeneratedProtocolMessageType('RulesList', (_message.Message,), {
  'DESCRIPTOR' : _RULESLIST,
  '__module__' : 'rules_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.RulesList)
  })
_sym_db.RegisterMessage(RulesList)

RuleBuilder = _reflection.GeneratedProtocolMessageType('RuleBuilder', (_message.Message,), {
  'DESCRIPTOR' : _RULEBUILDER,
  '__module__' : 'rules_pb2'
  # @@protoc_insertion_point(class_scope:lmao_pb.RuleBuilder)
  })
_sym_db.RegisterMessage(RuleBuilder)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z-github.com/NimbleBoxAI/nimblebox-lmao/lmao_pb'
  _LOGIC._serialized_start=88
  _LOGIC._serialized_end=247
  _INITRULEREQUEST._serialized_start=249
  _INITRULEREQUEST._serialized_end=358
  _RULE._serialized_start=361
  _RULE._serialized_end=612
  _RULESLIST._serialized_start=614
  _RULESLIST._serialized_end=697
  _RULEBUILDER._serialized_start=699
  _RULEBUILDER._serialized_end=785
# @@protoc_insertion_point(module_scope)



import os
import grpc
from typing import Union, Dict, List, Any

from nbox import utils as U
from nbox.auth import secret, NBX_JOB_TYPE, NBX_DEPLOY_TYPE
from nbox.utils import logger, lo
from nbox.relics import Relics

from google.protobuf.struct_pb2 import Struct

from nbox.lmao_v4.proto.project_pb2 import Project as ProjectProto
from nbox.lmao_v4.proto.tracker_pb2 import Tracker as TrackerProto, TrackerType, InitTrackerRequest, NBX as NBXType
from nbox.lmao_v4.proto.logs_pb2 import TrackerLogId
from nbox.lmao_v4.common import get_lmao_stub, get_tracker_log


class Tracker():
  """This is the generic class for any kind of Tracker, users can submodule this and leverage the existing methods"""
  def __init__(
      self,
      project_id: Union[str, ProjectProto],
      tracker_id: Union[str, TrackerProto] = "",
      live_tracker: bool = False,
      config: Dict[str, Any] = {},
      *,
      buffer_size: int = 10
    ):
    self.stub = get_lmao_stub()
    self.project_pb = project_id
    if type(project_id) != ProjectProto:
      self.project_pb = self.stub.GetProject(ProjectProto(id = project_id))
    
    # The big tracker switch case
    if tracker_id == "":
      logger.info("Initialising NBX-tracker")
      action = "Initialised"
      self.tracker_pb = self._init_tracker(
        project_id = self.project_pb.id,
        tracker_type = TrackerType.LIVE if live_tracker else TrackerType.EXPERIMENT,
        config = config
      )
    elif type(tracker_id) == TrackerProto:
      logger.info("Loaded NBX-tracker")
      action = "Loaded"
      self.tracker_pb = tracker_id
    elif isinstance(tracker_id, str):
      logger.info("Getting NBX-tracker")
      action = "Got"
      self.tracker_pb = self.stub.GetTracker(TrackerProto(project_id = self.project_pb.id, id=tracker_id))
    else:
      raise ValueError("Invalid tracker_id")
    
    logger.info(lo(
      f"{action} NBX-tracker",
      project = self.project_pb.id,
      tracker = self.tracker_pb.id,
      link = f"{secret.nbx_url}/workspace/{self.tracker_pb.workspace_id}/projects/{self.project_pb.id}"
    ))

    self.relic = Relics(id = self.project_pb.relic_id, prefix = self.tracker_pb.save_location)
    self.total_logged_elements = 0
    self.buffer_size = buffer_size
    self.pending_log_buffer = [] # this is for the rolling buffer when 

  def __repr__(self) -> str:
    return (f"    project: {self.project_pb.id}\n"
            f" tracker-id: {self.tracker_pb.id}\n"
            f"       link: {secret.nbx_url}/workspace/{self.tracker_pb.workspace_id}/projects/{self.project_pb.id}")

  def get_relic(self):
    """Get the underlying Relic for more advanced usage patterns."""
    return self.relic

  # before even thinking we should directly expose all the underlying rpc methods
  def _init_tracker(self, project_id, tracker_type: int, config: Dict[str, Any] = {}):
    # get the agent details
    agent_details = secret.get_agent_details()
    agent_type = NBXType.LOCAL
    if agent_details == NBX_JOB_TYPE:
      agent_type = NBXType.JOB
    elif agent_details == NBX_DEPLOY_TYPE:
      agent_type = NBXType.SERVING

    # get the struct for the config
    config_struct = Struct()
    config_struct.update(config)

    # initialise the tracker and return
    tracker: TrackerProto = self.stub.InitTracker(
      InitTrackerRequest(
        project_id = project_id,
        tracker_type = tracker_type,
        nbx_group_id = agent_details.group_id,
        nbx_instance_id = agent_details.instance_id,
        type = agent_type,
        config = config_struct
      )
    )
    return tracker
  
  def update_tracker_agent(self):
    agent_details = secret.get_agent_details()
    agent_type = NBXType.LOCAL
    if agent_details.nbx_type == NBX_JOB_TYPE:
      agent_type = NBXType.JOB
    elif agent_details.nbx_type == NBX_DEPLOY_TYPE:
      agent_type = NBXType.SERVING

    logger.info(f"Updating tracker agent: {agent_details.group_id} - {agent_details.instance_id}")
    self.tracker_pb.nbx_group_id = agent_details.group_id
    self.tracker_pb.nbx_instance_id = agent_details.instance_id
    self.tracker_pb.type = agent_type
    self.tracker_pb.update_keys.extend(("nbx_group_id", "nbx_instance_id", "type"))
    self.tracker_pb: TrackerProto = self.stub.UpdateTracker(self.tracker_pb)

  def log(self, log: Dict[str, Union[int, float, str]], *, log_id: str = "") -> str:
    if self.tracker_pb.status == TrackerProto.Status.COMPLETED:
      raise ValueError("Cannot log to a completed tracker")
    tracker_log = get_tracker_log(log)
    if log_id == "":
      U.is_valid_uuid(log_id)
      tracker_log.log_id = log_id
    tracker_log.project_id = self.tracker_pb.project_id
    tracker_log.tracker_id = self.tracker_pb.id
    self.pending_log_buffer.append(log)

    try:
      for log in self.pending_log_buffer:
        log_id: TrackerLogId = self.stub.PutTrackerLog(tracker_log)
        self.total_logged_elements += 1
    except grpc.RpcError as e:
      logger.warning(f"Failed to log: {log}")
      self.pending_log_buffer.append(log)
      return ""

    return log_id.log_id

  def end(self):
    """End the tracker to declare it complete. This will then trigger all the post end operations"""
    if self.tracker_pb.status == TrackerProto.Status.COMPLETED:
      raise ValueError("Cannot end a completed tracker")
    logger.info(f"Ending project ('{self.tracker_pb.project_id}') tracker: '{self.tracker_pb.id}'")
    self.tracker_pb.status = TrackerProto.Status.COMPLETED
    self.tracker_pb.update_keys.append("status")
    self.tracker_pb = self.stub.UpdateTracker(self.tracker_pb)

  def save_file(self, *files: List[str]):
    """
    Args:
      files: The list of files to save. This can be a list of files or a list of folders. If a folder is passed, all the files in the folder will be uploaded.
    """
    logger.info(f"Saving files: {files}")
    # manage all the complexity of getting the list of RelicFile
    all_files = []
    for folder_or_file in files:
      if os.path.isfile(folder_or_file):
        all_files.append(folder_or_file)
      elif os.path.isdir(folder_or_file):
        all_files.extend(U.get_files_in_folder(folder_or_file, abs_path=False))
      else:
        raise Exception(f"File or Folder not found: {folder_or_file}")

    # when we add something cool we can use this
    logger.debug(f"Storing {len(all_files)} files")
    relic = self.get_relic()
    for f in all_files:
      relic.put(f)
    return all_files

  def add_files(self, *files: List[str]):
    logger.warning("Tracker.add_files is deprecated, please use save_file instead")
    return self.save_file(*files)
  
  def delete_tracker(self):
    logger.info(f"Deleting tracker: {self.tracker_pb.id}")
    self.stub.DeleteTracker(self.tracker_pb)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
try:
  from gen.proto import logs_pb2 as logs__pb2
except ImportError:
  from nbox.lmao_v4.proto import logs_pb2 as logs__pb2
try:
  from gen.proto import project_pb2 as project__pb2
except ImportError:
  from nbox.lmao_v4.proto import project_pb2 as project__pb2
try:
  from gen.proto import rules_pb2 as rules__pb2
except ImportError:
  from nbox.lmao_v4.proto import rules_pb2 as rules__pb2
try:
  from gen.proto import tracker_pb2 as tracker__pb2
except ImportError:
  from nbox.lmao_v4.proto import tracker_pb2 as tracker__pb2


class LMAOStub(object):
    """This is the service definition for the LMAO server, this server will talk to the MongoDB in the backend.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.InitProject = channel.unary_unary(
                '/lmao_pb.LMAO/InitProject',
                request_serializer=project__pb2.InitProjectRequest.SerializeToString,
                response_deserializer=project__pb2.InitProjectResponse.FromString,
                )
        self.GetProject = channel.unary_unary(
                '/lmao_pb.LMAO/GetProject',
                request_serializer=project__pb2.Project.SerializeToString,
                response_deserializer=project__pb2.Project.FromString,
                )
        self.UpdateProject = channel.unary_unary(
                '/lmao_pb.LMAO/UpdateProject',
                request_serializer=project__pb2.Project.SerializeToString,
                response_deserializer=project__pb2.Project.FromString,
                )
        self.DeleteProject = channel.unary_unary(
                '/lmao_pb.LMAO/DeleteProject',
                request_serializer=project__pb2.Project.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.ListProjects = channel.unary_unary(
                '/lmao_pb.LMAO/ListProjects',
                request_serializer=project__pb2.ListProjectsRequest.SerializeToString,
                response_deserializer=project__pb2.ListProjectsResponse.FromString,
                )
        self.GetRuleBuilder = channel.unary_unary(
                '/lmao_pb.LMAO/GetRuleBuilder',
                request_serializer=rules__pb2.RuleBuilder.SerializeToString,
                response_deserializer=rules__pb2.RuleBuilder.FromString,
                )
        self.CreateRule = channel.unary_unary(
                '/lmao_pb.LMAO/CreateRule',
                request_serializer=rules__pb2.InitRuleRequest.SerializeToString,
                response_deserializer=rules__pb2.Rule.FromString,
                )
        self.GetRule = channel.unary_unary(
                '/lmao_pb.LMAO/GetRule',
                request_serializer=rules__pb2.Rule.SerializeToString,
                response_deserializer=rules__pb2.Rule.FromString,
                )
        self.UpdateRule = channel.unary_unary(
                '/lmao_pb.LMAO/UpdateRule',
                request_serializer=rules__pb2.Rule.SerializeToString,
                response_deserializer=rules__pb2.Rule.FromString,
                )
        self.DeleteRule = channel.unary_unary(
                '/lmao_pb.LMAO/DeleteRule',
                request_serializer=rules__pb2.Rule.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.ListRules = channel.unary_unary(
                '/lmao_pb.LMAO/ListRules',
                request_serializer=rules__pb2.RulesList.SerializeToString,
                response_deserializer=rules__pb2.RulesList.FromString,
                )
        self.InitTracker = channel.unary_unary(
                '/lmao_pb.LMAO/InitTracker',
                request_serializer=tracker__pb2.InitTrackerRequest.SerializeToString,
                response_deserializer=tracker__pb2.Tracker.FromString,
                )
        self.GetTracker = channel.unary_unary(
                '/lmao_pb.LMAO/GetTracker',
                request_serializer=tracker__pb2.Tracker.SerializeToString,
                response_deserializer=tracker__pb2.Tracker.FromString,
                )
        self.UpdateTracker = channel.unary_unary(
                '/lmao_pb.LMAO/UpdateTracker',
                request_serializer=tracker__pb2.Tracker.SerializeToString,
                response_deserializer=tracker__pb2.Tracker.FromString,
                )
        self.DeleteTracker = channel.unary_unary(
                '/lmao_pb.LMAO/DeleteTracker',
                request_serializer=tracker__pb2.Tracker.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.ListTrackers = channel.unary_unary(
                '/lmao_pb.LMAO/ListTrackers',
                request_serializer=tracker__pb2.ListTrackersRequest.SerializeToString,
                response_deserializer=tracker__pb2.ListTrackersResponse.FromString,
                )
        self.GetTrackerTable = channel.unary_unary(
                '/lmao_pb.LMAO/GetTrackerTable',
                request_serializer=logs__pb2.GetTrackerTableRequest.SerializeToString,
                response_deserializer=logs__pb2.TrackerTable.FromString,
                )
        self.PutTrackerLog = channel.unary_unary(
                '/lmao_pb.LMAO/PutTrackerLog',
                request_serializer=logs__pb2.TrackerLog.SerializeToString,
                response_deserializer=logs__pb2.TrackerLogId.FromString,
                )
        self.GetTrackerLogs = channel.unary_unary(
                '/lmao_pb.LMAO/GetTrackerLogs',
                request_serializer=logs__pb2.TrackerLogRequest.SerializeToString,
                response_deserializer=logs__pb2.TrackerLogResponse.FromString,
                )
        self.CreateDataset = channel.unary_unary(
                '/lmao_pb.LMAO/CreateDataset',
                request_serializer=logs__pb2.TrackerDatasetRequest.SerializeToString,
                response_deserializer=logs__pb2.TrackerDataset.FromString,
                )
        self.CreateDatasetStatus = channel.unary_unary(
                '/lmao_pb.LMAO/CreateDatasetStatus',
                request_serializer=logs__pb2.TrackerDataset.SerializeToString,
                response_deserializer=logs__pb2.TrackerDataset.FromString,
                )


class LMAOServicer(object):
    """This is the service definition for the LMAO server, this server will talk to the MongoDB in the backend.
    """

    def InitProject(self, request, context):
        """Project level information -> Full CRUDL on project 
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetProject(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateProject(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteProject(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListProjects(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRuleBuilder(self, request, context):
        """Rules based workflow -> Full CRUDL for rules 
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateRule(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRule(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateRule(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteRule(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListRules(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InitTracker(self, request, context):
        """Generic Tracking -> Full CRUDL on trackers 
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTracker(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateTracker(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteTracker(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListTrackers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTrackerTable(self, request, context):
        """things for logs
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PutTrackerLog(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTrackerLogs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateDataset(self, request, context):
        """
        This is the new service that we want to add 

        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateDatasetStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_LMAOServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'InitProject': grpc.unary_unary_rpc_method_handler(
                    servicer.InitProject,
                    request_deserializer=project__pb2.InitProjectRequest.FromString,
                    response_serializer=project__pb2.InitProjectResponse.SerializeToString,
            ),
            'GetProject': grpc.unary_unary_rpc_method_handler(
                    servicer.GetProject,
                    request_deserializer=project__pb2.Project.FromString,
                    response_serializer=project__pb2.Project.SerializeToString,
            ),
            'UpdateProject': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateProject,
                    request_deserializer=project__pb2.Project.FromString,
                    response_serializer=project__pb2.Project.SerializeToString,
            ),
            'DeleteProject': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteProject,
                    request_deserializer=project__pb2.Project.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'ListProjects': grpc.unary_unary_rpc_method_handler(
                    servicer.ListProjects,
                    request_deserializer=project__pb2.ListProjectsRequest.FromString,
                    response_serializer=project__pb2.ListProjectsResponse.SerializeToString,
            ),
            'GetRuleBuilder': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRuleBuilder,
                    request_deserializer=rules__pb2.RuleBuilder.FromString,
                    response_serializer=rules__pb2.RuleBuilder.SerializeToString,
            ),
            'CreateRule': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateRule,
                    request_deserializer=rules__pb2.InitRuleRequest.FromString,
                    response_serializer=rules__pb2.Rule.SerializeToString,
            ),
            'GetRule': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRule,
                    request_deserializer=rules__pb2.Rule.FromString,
                    response_serializer=rules__pb2.Rule.SerializeToString,
            ),
            'UpdateRule': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateRule,
                    request_deserializer=rules__pb2.Rule.FromString,
                    response_serializer=rules__pb2.Rule.SerializeToString,
            ),
            'DeleteRule': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteRule,
                    request_deserializer=rules__pb2.Rule.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'ListRules': grpc.unary_unary_rpc_method_handler(
                    servicer.ListRules,
                    request_deserializer=rules__pb2.RulesList.FromString,
                    response_serializer=rules__pb2.RulesList.SerializeToString,
            ),
            'InitTracker': grpc.unary_unary_rpc_method_handler(
                    servicer.InitTracker,
                    request_deserializer=tracker__pb2.InitTrackerRequest.FromString,
                    response_serializer=tracker__pb2.Tracker.SerializeToString,
            ),
            'GetTracker': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTracker,
                    request_deserializer=tracker__pb2.Tracker.FromString,
                    response_serializer=tracker__pb2.Tracker.SerializeToString,
            ),
            'UpdateTracker': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateTracker,
                    request_deserializer=tracker__pb2.Tracker.FromString,
                    response_serializer=tracker__pb2.Tracker.SerializeToString,
            ),
            'DeleteTracker': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteTracker,
                    request_deserializer=tracker__pb2.Tracker.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'ListTrackers': grpc.unary_unary_rpc_method_handler(
                    servicer.ListTrackers,
                    request_deserializer=tracker__pb2.ListTrackersRequest.FromString,
                    response_serializer=tracker__pb2.ListTrackersResponse.SerializeToString,
            ),
            'GetTrackerTable': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTrackerTable,
                    request_deserializer=logs__pb2.GetTrackerTableRequest.FromString,
                    response_serializer=logs__pb2.TrackerTable.SerializeToString,
            ),
            'PutTrackerLog': grpc.unary_unary_rpc_method_handler(
                    servicer.PutTrackerLog,
                    request_deserializer=logs__pb2.TrackerLog.FromString,
                    response_serializer=logs__pb2.TrackerLogId.SerializeToString,
            ),
            'GetTrackerLogs': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTrackerLogs,
                    request_deserializer=logs__pb2.TrackerLogRequest.FromString,
                    response_serializer=logs__pb2.TrackerLogResponse.SerializeToString,
            ),
            'CreateDataset': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateDataset,
                    request_deserializer=logs__pb2.TrackerDatasetRequest.FromString,
                    response_serializer=logs__pb2.TrackerDataset.SerializeToString,
            ),
            'CreateDatasetStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateDatasetStatus,
                    request_deserializer=logs__pb2.TrackerDataset.FromString,
                    response_serializer=logs__pb2.TrackerDataset.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'lmao_pb.LMAO', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class LMAO(object):
    """This is the service definition for the LMAO server, this server will talk to the MongoDB in the backend.
    """

    @staticmethod
    def InitProject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/InitProject',
            project__pb2.InitProjectRequest.SerializeToString,
            project__pb2.InitProjectResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetProject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/GetProject',
            project__pb2.Project.SerializeToString,
            project__pb2.Project.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateProject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/UpdateProject',
            project__pb2.Project.SerializeToString,
            project__pb2.Project.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteProject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/DeleteProject',
            project__pb2.Project.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListProjects(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/ListProjects',
            project__pb2.ListProjectsRequest.SerializeToString,
            project__pb2.ListProjectsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetRuleBuilder(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/GetRuleBuilder',
            rules__pb2.RuleBuilder.SerializeToString,
            rules__pb2.RuleBuilder.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateRule(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/CreateRule',
            rules__pb2.InitRuleRequest.SerializeToString,
            rules__pb2.Rule.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetRule(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/GetRule',
            rules__pb2.Rule.SerializeToString,
            rules__pb2.Rule.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateRule(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/UpdateRule',
            rules__pb2.Rule.SerializeToString,
            rules__pb2.Rule.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteRule(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/DeleteRule',
            rules__pb2.Rule.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListRules(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/ListRules',
            rules__pb2.RulesList.SerializeToString,
            rules__pb2.RulesList.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def InitTracker(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/InitTracker',
            tracker__pb2.InitTrackerRequest.SerializeToString,
            tracker__pb2.Tracker.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTracker(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/GetTracker',
            tracker__pb2.Tracker.SerializeToString,
            tracker__pb2.Tracker.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateTracker(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/UpdateTracker',
            tracker__pb2.Tracker.SerializeToString,
            tracker__pb2.Tracker.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteTracker(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/DeleteTracker',
            tracker__pb2.Tracker.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListTrackers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/ListTrackers',
            tracker__pb2.ListTrackersRequest.SerializeToString,
            tracker__pb2.ListTrackersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTrackerTable(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/GetTrackerTable',
            logs__pb2.GetTrackerTableRequest.SerializeToString,
            logs__pb2.TrackerTable.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PutTrackerLog(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/PutTrackerLog',
            logs__pb2.TrackerLog.SerializeToString,
            logs__pb2.TrackerLogId.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTrackerLogs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/GetTrackerLogs',
            logs__pb2.TrackerLogRequest.SerializeToString,
            logs__pb2.TrackerLogResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateDataset(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/CreateDataset',
            logs__pb2.TrackerDatasetRequest.SerializeToString,
            logs__pb2.TrackerDataset.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateDatasetStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lmao_pb.LMAO/CreateDatasetStatus',
            logs__pb2.TrackerDataset.SerializeToString,
            logs__pb2.TrackerDataset.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import os
import sys
import ast
import shlex
import zipfile
from subprocess import Popen
from typing import Tuple, Dict

from nbox import utils as U
from nbox.auth import secret
from nbox.utils import logger, lo
from nbox.relics import Relics
from nbox import messages as mpb
from nbox.init import nbox_ws_v1
from nbox.jd_core import Job, Serve, upload_job_folder

from nbox.nbxlib.astea import Astea

from nbox.hyperloop.common.common_pb2 import Resource
from nbox.lmao_v4 import get_lmao_stub, get_git_details, LMAO_RM_PREFIX, ExperimentConfig, Tracker
from nbox.lmao_v4.proto.lmao_service_pb2_grpc import LMAOStub
from nbox.lmao_v4.proto import tracker_pb2 as t_pb
from nbox.lmao_v4.proto import project_pb2 as p_pb


def evaluate_unary_op(node):
  if isinstance(node.op, ast.USub):  # Unary negation
    return -evaluate_node(node.operand)
  elif isinstance(node.op, ast.UAdd):  # Unary positive
    return evaluate_node(node.operand)
  elif isinstance(node.op, ast.Not):  # Logical not
    return not evaluate_node(node.operand)
  else:
    raise ValueError("Unsupported unary operator")

def evaluate_node(node):
  if isinstance(node, ast.Constant):
    return node.value
  elif isinstance(node, ast.UnaryOp):
    return evaluate_unary_op(node)
  else:
    raise ValueError("Unsupported node type")

def _parse_job_code(init_path: str, run_kwargs) -> Tuple[str, Dict]:
  # analyse the input file and function, extract the types and build the run_kwargs
  fn_file, fn_name = init_path.split(":")
  if not os.path.exists(fn_file + ".py"):
    raise Exception(f"File {fn_file} does not exist")
  init_folder, fn_file = os.path.split(fn_file + ".py")
  init_folder = init_folder or "."
  tea = Astea(fn_file)
  items = tea.find(fn_name)
  if not len(items):
    raise Exception(f"Function {fn_name} not found in {fn_file}")
  fn = items[0]
  defaults = fn.node.args.defaults
  args = fn.node.args.args
  _default_kwarg = {}
  for k,v in zip(args[::-1], defaults[::-1]):
    _k = k.arg
    if type(v) == ast.UnaryOp:
      logger.warning(f"Default value for var '{_k}' is a UnaryOp, cannot validate the type")
      try:
        evaluate_node(v)
      except:
        raise Exception(f"Default value for var '{_k}' is not a primitive type [int, float, str, bool, None] but {type(v)}")
      _default_kwarg[_k] = evaluate_node(v)
    elif not type(v) == ast.Constant:
      raise Exception(f"Default value for var '{_k}' is not a primitive type [int, float, str, bool, None] but {type(v)}")
    else:
      try:
        _default_kwarg[_k] = v.value
      except:
        raise Exception(f"Default value for var '{_k}' is not a primitive type [int, float, str, bool, None] but {type(v)}")
  args_set = set([x.arg for x in args])
  args_na = args_set - set(_default_kwarg.keys())
  args_not_passed = set(args_na) - set(run_kwargs.keys())
  if len(args_not_passed):
    raise Exception(f"Following arguments are not passed: {args_not_passed} but required, this can cause errors during execution")
  extra_args = set(run_kwargs.keys()) - args_set
  if len(extra_args) and not fn.node.args.kwarg:
    raise Exception(f"Following arguments are passed but not consumed by {fn_name}: {extra_args}")
  final_dict = {}

  # left merge the two dicts: default_kwargs <- run_kwargs
  merged = {**_default_kwarg, **run_kwargs}
  for k,v in merged.items():
    if type(v) == type(None):
      final_dict[k] = run_kwargs.get(k, v)
    else:
      final_dict[k] = type(v)(run_kwargs.get(k, v))
  return final_dict, init_folder



### ---------------
# this will be eventually merged with the project in the root scope


_SUPPORTED_SERVER_TYPES = ["fastapi"]

class ProjectState:
  project_id: str = ""
  experiment_id: str = ""
  serving_id: str = ""

  def data():
    return {k:getattr(ProjectState, k) for k in ProjectState.__dict__ if not k.startswith("__") and k != "data"}


# aka Project_v4
# TODO: @yashbonde: all the jobs are Job gRPC stubs
# TODO: @yashbonde: all the experiments are Experiment gRPC stubs
class Project:
  def __init__(self, id: str = ""):
    id = id or ProjectState.project_id
    if not id:
      raise ValueError("Project ID is not set")
    logger.info(f"Connecting to Project: {id}")

    self.stub = nbox_ws_v1.projects.u(id)
    self.data = self.stub()
    self.lmao_stub: LMAOStub = get_lmao_stub()
    _p = p_pb.Project(
      id = id
    )
    self.project_pb = self.lmao_stub.GetProject(_p)
    if self.project_pb is None:
      raise ValueError("Could not connect to Monitoring backend.")
    self.artifact = Relics(id = self.project_pb.relic_id)
    self.workspace_id = secret.workspace_id

  def __repr__(self) -> str:
    return f"Project({self.project_pb.id})"

  # things about the project as a resource

  def list_trackers(self, live: bool = False, status: str = t_pb.Tracker.Status.UNSET_STATUS) -> t_pb.ListTrackersResponse:
    req = t_pb.ListTrackersRequest(
      project_id = self.project_pb.id,
      status = status,
      tracker_type = t_pb.TrackerType.LIVE if live else t_pb.TrackerType.EXPERIMENT,
    )
    return self.lmao_stub.ListTrackers(req)

  @property
  def metadata(self):
    """A NimbleBox project is a very large entity and its components are in multiple places.
    `metadata` is a dictionary that contains all the information about the project."""
    return {"details": self.data, "lmao": mpb.MessageToDict(self.project_pb, including_default_value_fields=True)}

  def put_settings(self, project_name: str = "", project_description: str = ""):
    self.stub(
      "put",
      project_name = project_name or self.data["project_name"],
      project_description = project_description or self.data["project_description"]
    )
    self.data["project_name"] = project_name or self.data["project_name"]
    self.data["project_description"] = project_description or self.data["project_description"]

  # some getters for underlying objects

  def get_lmao_stub(self) -> LMAOStub:
    return self.lmao_stub
  
  def get_artifact(self) -> Relics:
    return self.artifact

  def get_job_id(self) -> str:
    return self.data["job_list"][0]

  def get_deployment_id(self) -> str:
    return self.data["deployment_list"][0]

  def get_tracker(self, tracker_type: int, tracker_id: str, config) -> Tracker:
    return Tracker(
      project_id = self.project_pb,
      tracker_id = tracker_id,
      config = config,
      live_tracker = tracker_type == t_pb.TrackerType.LIVE,
    )

  def get_exp_tracker(
      self,
      experiment_id: str = "",
      metadata = {},
      update_agent: bool = False,
    ) -> Tracker:
    if not experiment_id and ProjectState.experiment_id:
      experiment_id = ProjectState.experiment_id
    tracker = self.get_tracker(
      tracker_type = t_pb.TrackerType.EXPERIMENT,
      tracker_id = experiment_id,
      config = metadata
    )
    if update_agent:
      tracker.update_tracker_agent()
    return tracker

  def get_live_tracker(self, serving_id: str = "", metadata = {}) -> Tracker:
    return self.get_tracker(
      tracker_type = t_pb.TrackerType.LIVE,
      tracker_id = serving_id,
      config = metadata
    )

  # Some big things primarily built for the cli but can be reused as APIs as well

  def run(
    self,
    init_path: str,

    # all the arguments for the git files
    untracked: bool = False,
    upload_git: bool = False,
    untracked_no_limit: bool = False,

    # all the things for resources
    resource_cpu: str = "",
    resource_memory: str = "",
    resource_disk_size: str = "",
    resource_gpu: str = "",
    resource_gpu_count: str = "",
    resource_max_retries: int = 0,

    # any other things to pass to the function / class being called
    **run_kwargs,
  ):
    """Upload and register a new run for a NBX-LMAO project.

    Args:
      init_path (str): This can be a path to a `folder` or can be optionally of the structure `fp:fn` where `fp` is the path to the file and `fn` is the function name.
      untracked (bool, optional): If True, then untracked files below 1MB will also be zipped and uploaded. Defaults to False.
      untracked_no_limit (bool, optional): If True, then all untracked files will also be zipped and uploaded. Defaults to False.
      resource_cpu (str, optional): The CPU resource to allocate to the run. It can be found on your NimbleBox dashboard.
      resource_memory (str, optional): The memory resource to allocate to the run. It can be found on your NimbleBox dashboard.
      resource_disk_size (str, optional): The disk size resource to allocate to the run. It can be found on your NimbleBox dashboard.
      resource_gpu (str, optional): The GPU resource to use. It can be found on your NimbleBox dashboard.
      resource_gpu_count (str, optional): Number of GPUs allocated to the experiment. It can be found on your NimbleBox dashboard.
      resource_max_retries (int, optional): The maximum number of retries for a run. It can be found on your NimbleBox dashboard.
      **run_kwargs: These are the kwargs that will be passed to your function.
    """
    # fix data type conversion caused by the CLI
    resource_cpu = str(resource_cpu)
    resource_memory = str(resource_memory)
    resource_disk_size = str(resource_disk_size)
    resource_gpu = str(resource_gpu)
    resource_gpu_count = str(resource_gpu_count)
    resource_max_retries = int(resource_max_retries)

    # reconstruct the entire CLI command so we can show it in the UI
    reconstructed_cli_comm = (
      f"nbx projects --id '{self.project_pb.id}' - run '{init_path}'" +
      (f" --resource_cpu '{resource_cpu}'" if resource_cpu else "") +
      (f" --resource_memory '{resource_memory}'" if resource_memory else "") +
      (f" --resource_disk_size '{resource_disk_size}'" if resource_disk_size else "") +
      (f" --resource_gpu '{resource_gpu}'" if resource_gpu else "") +
      (f" --resource_gpu_count '{resource_gpu_count}'" if resource_gpu_count else "") +
      (f" --resource_max_retries {resource_max_retries}" if resource_max_retries else "")
    )
    if untracked_no_limit and not untracked:
      raise ValueError("Cannot set untracked_no_limit to True if untracked is False")
    if untracked:
      reconstructed_cli_comm += " --untracked"
    if untracked_no_limit:
      reconstructed_cli_comm += " --untracked_no_limit"
    for k, v in run_kwargs.items():
      if type(v) == bool:
        reconstructed_cli_comm += f" --{k}"
      elif type(v) == str:
        reconstructed_cli_comm += f" --{k} '{v}'"
      elif isinstance(v, (int, float)):
        reconstructed_cli_comm += f" --{k} {v}"
      elif isinstance(v, (list, tuple)):
        reconstructed_cli_comm += f" --{k} \"[{v}]\""
      else:
        reconstructed_cli_comm += f" --{k} '{v}'"
    logger.debug(f"command: {reconstructed_cli_comm}")

    # get the job
    job = Job(job_id = self.get_job_id())
    job.job_proto.resource.MergeFrom(Resource(
      cpu = resource_cpu,
      memory = resource_memory,
      disk_size = resource_disk_size,
      gpu = resource_gpu,
      gpu_count = resource_gpu_count,
      max_retries = resource_max_retries,
    ))

    # parse the code file + get git details, these two steps are unique to LMAO Run
    run_kwargs, init_folder = _parse_job_code(init_path = init_path, run_kwargs = run_kwargs) 
    if os.path.exists(U.join(init_folder, ".git")):
      git_det = get_git_details(init_folder)
    else:
      git_det = {}

    # initialise the run with the big object and get the experiment_id
    exp_config = ExperimentConfig(
      run_kwargs = run_kwargs,
      git = git_det,
      resource = job.job_proto.resource,
      cli_comm = reconstructed_cli_comm,
      save_to_relic = True,
      enable_system_monitoring = False,
    ).to_dict()
    logger.debug(lo("Experiment Config", **exp_config))
    exp_tracker = self.get_exp_tracker(metadata = exp_config)
    tracker_pb = exp_tracker.tracker_pb
    logger.info(f"Created new tracker ID: {tracker_pb.id}")

    # create a git patch and upload it to relics, this is unique to LMAO Run
    if git_det and upload_git:
      _zf = U.join(init_folder, "untracked.zip")
      zip_file = zipfile.ZipFile(_zf, "w", zipfile.ZIP_DEFLATED)
      patch_file = U.join(init_folder, "nbx_auto_patch.diff")
      f = open(patch_file, "w")
      Popen(shlex.split(f"git diff {' '.join(git_det['uncommited_files'])}"), stdout=f, stderr=sys.stderr).wait()
      f.close()
      zip_file.write(patch_file, arcname = patch_file)
      if untracked:
        untracked_files = git_det["untracked_files"] # what to do with these?
        if untracked_files:
          # see if any file is larger than 10MB and if so, warn the user
          for f in untracked_files:
            if os.path.getsize(f) > 1e7 and not untracked_no_limit:
              logger.warning(f"File: {f} is larger than 10MB and will not be available in sync")
              logger.warning("  Fix: use git to track small files, avoid large files")
              logger.warning("  Fix: nbox.Relics can be used to store large files")
              logger.warning("  Fix: use --untracked-no-limit to upload all files")
              continue
            zip_file.write(f, arcname = f)

      # upload the patch file to relics
      self.artifact.put_to(_zf, f"{tracker_pb.id}/git.zip")
      os.remove(_zf)
      os.remove(patch_file)

    # tell the server that this run is being scheduled so atleast the information is visible on the dashboard
    upload_job_folder(
      method = "job",
      init_folder = init_path,
      project_id = self.project_pb.id,

      # pass along the resource requirements
      resource_cpu = resource_cpu,
      resource_memory = resource_memory,
      resource_disk_size = resource_disk_size,
      resource_gpu = resource_gpu,
      resource_gpu_count = resource_gpu_count,
      resource_max_retries = resource_max_retries,
    )

    # now trigger the experiment
    fp = f"{tracker_pb.project_id}/{tracker_pb.id}"
    tag = f"{LMAO_RM_PREFIX}{fp}"
    logger.info(f"Running job '{job.name}' ({job.id}) with tag: {tag}")
    job.trigger(tag)

    # finally print the location of the run where the users can track this
    logger.info(f"Run location: {secret.nbx_url}/workspace/{secret.workspace_id}/projects/{self.project_pb.id}#Experiments")

  def serve(
    self,
    init_path: str,
    server_type: str = "fastapi",

    # all the things for resources
    resource_cpu: str = "",
    resource_memory: str = "",
    resource_disk_size: str = "",
    resource_gpu: str = "",
    resource_gpu_count: str = "",
    resource_max_retries: int = 0,
  ):
    raise NotImplementedError("This is not implemented yet")
    if server_type not in _SUPPORTED_SERVER_TYPES:
      raise ValueError(f"server_type must be one of {_SUPPORTED_SERVER_TYPES}")
